"""
zmp.py

Functions associated with zmp inversion
"""

import psi4
import numpy as np
from functools import reduce
rng = np.random.default_rng(12345)

psi4.core.be_quiet()
eps = np.finfo(float).eps

class Random():
    def random(self, 
            max_iter=100, 
            density_tol= psi4.core.get_option("SCF", "D_CONVERGENCE"), 
            mixing    = 1e-4,
            print_scf = False, 
            ):

        """
        Performs random inversion according to:
        
        Using random numbers to obtain Kohn-Sham potential for a given density

        Additional DIIS algorithms obtained from:
        2) 'Psi4NumPy: An interactive quantum chemistry programming environment 
        for reference implementations and rapid development.' by 
        Daniel G.A. Smith and others. 
        https://doi.org/10.1021/acs.jctc.8b00286

        Functionals that drive the SCF procedure are obtained from:
        https://doi.org/10.1002/qua.26400

        """

        self.shift      = 0.001
        self.diis_space = 100
        self.mixing     = mixing

        print("\nRunning Random Inversion:")
        self.random_scf(max_iter, print_scf=print_scf, D_conv=density_tol)

    def random_scf(self, maxiter, print_scf,D_conv):
        """
        Performs scf cycle
        Parameters
        ----------
        maxiter: integer, opt
            Maximum number of iterations for SCF calculation
        D_conv: float
            Convergence parameter for density and DIIS error. 
            Default is Psi4's Density convergence parameter: 1e-06
        """

        #Checks if there is dft grid available. 
        if hasattr(self.wfn, 'V_potential()') == False:
            ref = "RV" if self.ref == 1 else "UV"
            functional = psi4.driver.dft.build_superfunctional("SVWN", restricted=True if self.ref==1 else False)[0]
            Vpot = psi4.core.VBase.build(self.basis, functional, ref)
            Vpot.initialize()
            self.vpot = Vpot
        else:
            self.vpot = self.wfn.V_potential()

        D0 = self.on_grid_density(grid=None, Da=self.nt[0], Db=self.nt[1],  vpot=self.vpot)

        self.vxc_additive = np.zeros_like(D0)
        vc_old = np.zeros((self.nbf, self.nbf))
        vc     = np.zeros((self.nbf, self.nbf))
        Da = np.zeros((self.nbf, self.nbf))
        Db = np.zeros((self.nbf, self.nbf))
        D_old  = np.zeros_like(Da)


        #------------->  Iterating over lambdas:

        # Trial & Residual Vector Lists
        state_vectors_a, state_vectors_b = [], []
        error_vectors_a, error_vectors_b = [], []

        for SCF_ITER in range(1,maxiter):

        #------------->  Generate Fock Matrix:
            vc = self.generate_random_functional(Da, Db)

            #Potential mixing
            vc_fock =  (1-self.mixing) * vc_old + (self.mixing) * (vc)

            #Equation 10 of Reference (1). Level shift. 
            Fa = self.T + self.V + self.va + vc_fock
            # Fa += (self.S2 - reduce(np.dot, (self.S2, Da, self.S2))) * self.shift

            # if self.ref == 2:
            #     Fb = self.T + self.V + self.vb + vc_fock
            #     Fb += (self.S2 - reduce(np.dot, (self.S2, Db, self.S2))) * self.shift
            

        #------------->  DIIS:
            if SCF_ITER > 1: 
                #Construct the AO gradient
                # r = (A(FDS - SDF)A)_mu_nu
                grad_a = self.A.dot(Fa.dot(Da).dot(self.S2) - self.S2.dot(Da).dot(Fa)).dot(self.A)
                grad_a[np.abs(grad_a) < eps] = 0.0

                if SCF_ITER < self.diis_space:
                    state_vectors_a.append(Fa.copy())
                    error_vectors_a.append(grad_a.copy())
                else:
                    state_vectors_a.append(Fa.copy())
                    error_vectors_a.append(grad_a.copy())
                    del state_vectors_a[-1]
                    del error_vectors_a[-1]

                #Build inner product of error vectors
                Bdim = len(state_vectors_a)
                Ba = np.empty((Bdim + 1, Bdim + 1))
                Ba[-1, :] = -1
                Ba[:, -1] = -1
                Ba[-1, -1] = 0
                Bb = Ba.copy()

                for i in range(len(state_vectors_a)):
                    for j in range(len(state_vectors_a)):
                        Ba[i,j] = np.einsum('ij,ij->', error_vectors_a[i], error_vectors_a[j])

                #Build almost zeros matrix to generate inverse. 
                RHS = np.zeros(Bdim + 1)
                RHS[-1] = -1

                #Find coefficient matrix:
                Ca = np.linalg.solve(Ba, RHS.copy())
                Ca[np.abs(Ca) < eps] = 0.0 

                #Generate new fock Matrix:
                Fa = np.zeros_like(Fa)
                for i in range(Ca.shape[0] - 1):
                    Fa += Ca[i] * state_vectors_a[i]

                diis_error_a = np.max(np.abs(error_vectors_a[-1]))
                if self.ref ==  1:
                    Fb = Fa.copy()
                    diis_error = 2 * diis_error_a

                else:
                    grad_b = self.A.dot(Fb.dot(Db).dot(self.S2) - self.S2.dot(Db).dot(Fb)).dot(self.A)
                    grad_b[np.abs(grad_b) < eps] = 0.0
                    
                    if SCF_ITER < self.diis_space:
                        state_vectors_b.append(Fb.copy())
                        error_vectors_b.append(grad_b.copy())
                    else:
                        state_vectors_b.append(Fb.copy())
                        error_vectors_b.append(grad_b.copy())
                        del state_vectors_b[-1]
                        del error_vectors_b[-1]

                    for i in range(len(state_vectors_b)):
                        for j in range(len(state_vectors_b)):
                            Bb[i,j] = np.einsum('ij,ij->', error_vectors_b[i], error_vectors_b[j])

                    diis_error_b = np.max(np.abs(error_vectors_b[-1]))
                    diis_error = diis_error_a + diis_error_b

                    Cb = np.linalg.solve(Bb, RHS.copy())
                    Cb[np.abs(Cb) < eps] = 0.0 

                    Fb = np.zeros_like(Fb)
                    for i in range(Cb.shape[0] - 1):
                        Fb += Cb[i] * state_vectors_b[i]

            else:
                diis_error = 1.0
            
        #------------->  Diagonalization | Check convergence:

            Ca, Cocca, Da, eigs_a = self.diagonalize(Fa, self.nalpha)
            if self.ref == 2:
                Cb, Coccb, Db, eigs_b = self.diagonalize(Fb, self.nbeta)
            else: 
                Cb, Coccb, Db, eigs_b = Ca.copy(), Cocca.copy(), Da.copy(), eigs_a.copy()

            ddm = D_old - Da
            vc_old = vc 
            D_old = Da
            derror = np.max(np.abs(ddm))

            density_current = self.on_grid_density(grid=None, Da=Da, Db=Db, vpot=self.vpot)
            grid_diff = np.max(np.abs(D0 - density_current))

            #Uncomment to debug internal SCF
            if print_scf:
                if np.mod(SCF_ITER, 1) == 0:  
                    print(f"Iteration: {SCF_ITER:3d} | Conv Error: {derror:10.10e} | DIIS Error:  | DD: {np.max(grid_diff)}")
            
            # if abs(derror) < D_conv: #and abs(diis_error) < D_conv:
            #     self.vc = vc
            #     break
            if SCF_ITER == maxiter - 1:
                raise ValueError("Maximum Number of SCF cycles reached. Try different settings.")

            density_current = self.on_grid_density(grid=None, Da=Da, Db=Db, vpot=self.vpot)
            grid_diff = np.max(np.abs(D0 - density_current))


    def generate_random_functional(self, Da, Db):
        """
        Generates S_n Functional as described by:
        Using random numbers to obtain Kohn-Sham potential for a given density
        Kumar + Harbola
        """

        # Obtain target density on dft_grid 
        D0 = self.on_grid_density(grid=None, Da=self.nt[0], Db=self.nt[1],  vpot=self.vpot)
        D_current = self.on_grid_density(grid=None, Da=Da, Db=Db,  vpot=self.vpot)
        target_density = D0
        current_density = D_current
        diff = current_density - target_density
        max_diff = np.max(diff)

        # Components for equation 12
        vc = np.empty_like(target_density)
        for i in range(diff.shape[0]):
                if current_density[i] < target_density[i]:
                    vc[i] =   max_diff * rng.random()
                else:
                    vc[i] = - max_diff * rng.random()

        self.vxc_additive += vc
        vc_nm = self.dft_grid_to_fock(vc, self.vpot)

        return vc_nm
        



"""
Provides interface n2v interface to Psi4
"""


from .engine import Engine
import numpy as np
from opt_einsum import contract

try:
    import psi4
    psi4.set_options({"save_jk" : True})
    has_psi4 = True
except ImportError:
    has_psi4 = False
    
if has_psi4:
    from ..grid import Psi4Grider
    class Psi4Engine(Engine):
        """
        Psi4 Engine Class
        """

        def set_system(self, molecule, basis, ref='1', pbs='same', wfn=None):
            """
            Initializes geometry and basis infromation

            Parameters
            ----------
            molecule: psi4.core.Molecule
                Molecule of the system  used
            basis: str
                Basis set of calculation
            ref: int
                Reference: Restricted   -> 1
                           Unrestricted -> 2
            pbs: str
                Basis set of potential used
            wfn : psi4.core.{RHF, UHF, RKS, UKS, Wavefunction, CCWavefuncion...}
                Psi4 wavefunction object
            """
            self.mol = molecule
            
            #Assert units are in bohr
            # units = self.mol.to_schema(dtype='psi4')['units']
            # if units != "Bohr":
            #     raise ValueError("Units need to be set in Bohr")
            self.basis_str = basis
            self.ref = ref 
            self.pbs = pbs
            self.pbs_str   = basis if pbs == 'same' else pbs

            self.nalpha = wfn.nalpha()
            self.nbeta = wfn.nbeta()

            self.wfn = wfn

        def initialize(self):
            """
            Initializes basic objects required for the Psi4Engine
            """
            self.basis = psi4.core.BasisSet.build( self.mol, key='BASIS', target=self.basis_str)
            self.pbs   = psi4.core.BasisSet.build( self.mol, key='BASIS', target=self.pbs_str)

            self.nbf   = self.basis.nbf()
            self.npbs  = self.pbs.nbf()

            self.mints = psi4.core.MintsHelper( self.basis )
            self.jk    = self.generate_jk()

            self.grid = Psi4Grider(self.mol, self.basis, self.ref)

        def get_T(self):
            """Kinetic Potential in ao basis"""
            return np.array( self.mints.ao_kinetic() )

        def get_Tpbas(self):
            """Kinetic Potential in pbs"""
            return np.array( self.mints.ao_kinetic(self.pbs, self.pbs) )

        def get_V(self):
            """External potential in ao basis"""
            return np.array( self.mints.ao_potential() )

        def get_A(self):
            """Inverse squared root of S matrix"""
            A = self.mints.ao_overlap()
            A.power( -0.5, 1e-16 )
            return np.array( A )

        def get_S(self):
            """Overlap matrix in AO basis"""
            return np.array( self.mints.ao_overlap() )

        def get_S3(self):
            """3 Orbitals Overlap matrix in AO basis"""
            return np.array( self.mints.ao_3coverlap(self.basis,self.basis,self.pbs) )

        def get_S4(self):
            """
            Calculates four overlap integral with Density Fitting method.
            S4_{ijkl} = \int dr \phi_i(r)*\phi_j(r)*\phi_k(r)*\phi_l(r)
            Parameters
            ----------
            wfn: psi4.core.Wavefunction
                Wavefunction object of moleculep
            Return
            ------
            S4
            """

            print(f"4-AO-Overlap tensor will take about {self.nbf **4 / 8 * 1e-9:f} GB.")

            aux_basis = psi4.core.BasisSet.build(self.mol, "DF_BASIS_SCF", "", "JKFIT", self.basis_str)
            S_Pmn = np.squeeze(self.mints.ao_3coverlap(aux_basis, self.basis, self.basis ))
            S_PQ = np.array(self.mints.ao_overlap(aux_basis, aux_basis))
            S_PQinv = np.linalg.pinv(S_PQ, rcond=1e-9)
            S4 = contract('Pmn,PQ,Qrs->mnrs', S_Pmn, S_PQinv, S_Pmn)
            return S4

        def generate_jk(self, gen_K=False):
            """
            Creates jk object for generation of Coulomb and Exchange matrices
            1.0e9 B -> 1.0 GB
            """
            jk = psi4.core.JK.build(self.basis)
            memory = int(jk.memory_estimate() * 1.1)
            jk.set_memory(int(memory)) 
            jk.set_do_K(gen_K)
            jk.initialize()

            return jk

        def compute_hartree(self, Cocc_a, Cocc_b):
            """
            Generates Coulomb and Exchange matrices from occupied orbitals
            """
            Cocc_a = psi4.core.Matrix.from_array(Cocc_a)
            Cocc_b = psi4.core.Matrix.from_array(Cocc_b)
            self.jk.C_left_add(Cocc_a)
            self.jk.C_left_add(Cocc_b) 
            self.jk.compute()
            self.jk.C_clear()
            J = (np.array(self.jk.J()[0]), np.array(self.jk.J()[1]))
            return J

        def hartree_NO(self, Dta):
            """
            Computes Hartree potential in AO basis from Natural Orbitals
            """

            if self.wfn is None:
                raise ValueError('Please provide a wfn object to the Inverter, i.e., Inverter.eng = wfn')

            if type(self.wfn) == psi4.core.CCWavefunction:
                C_NO    = psi4.core.Matrix(self.nbf, self.nbf)
                eigs_NO = psi4.core.Vector(self.nbf)
                self.wfn.Da().diagonalize( C_NO, eigs_NO, psi4.core.DiagonalizeOrder.Descending )
                occ = np.sqrt( np.array(eigs_NO) )
                new_CA = occ * np.array(C_NO)
                assert np.allclose( new_CA @ new_CA.T, Dta )
                if self.ref == 1:
                    new_CB = np.copy( new_CA )
                else:
                    self.wfn.Db().diagonalize( C_NO, eigs_NO, psi4.core.DiagonalizeOrder.Descending )
                    occ_b = np.sqrt( np.array( eigs_NO ) )
                    new_CB = occ_b * np.array( C_NO )
                J0 = self.compute_hartree(new_CA, new_CB)
                return J0

        def run_single_point(self, mol, basis, method):
            """
            Run a standard energy calculation
            """

            wfn_temp = psi4.energy(init+"/" + self.basis_str, 
                                molecule=self.mol, 
                                return_wfn=True)[1]

            if self.ref == 1:
                D = np.array(wfn_temp.Da()) + np.array(wfn_temp.Db())
                C = np.array(wfn_temp.Ca())
                e = np.array(wfn_temp.epsilon_a())
                
            else:
                D = np.stack( (np.array(wfn_temp.Da()), np.array(wfn_temp.Db())) )
                C = np.stack( (np.array(wfn_temp.Ca()), np.array(wfn_temp.Cb())) )
                e = np.stack( (np.array(wfn_temp.epsilon_a()), np.array(wfn_temp.epsilon_b())) )

            return D, C, e
            


"""
Provides interface n2v interface to PySCF
"""

import numpy as np
import scipy
from opt_einsum import contract

from .engine import Engine

try:
    from pyscf import gto, dft
    has_pyscf = True
except ImportError:
    has_pyscf = False

if has_pyscf:
    from ..grid import PySCFGrider
    class PySCFEngine(Engine):
        """
        PySCF Engine
        """

        def set_system(self, molecule, basis, ref=1, pbs='same'):
            """
            Stores basic information from a PySCF calculation

            Parameters
            ----------
            
            mol: pyscf.gto.mole.Mole
                Pyscf molecule object
            basis: str
                Basis set for calculation
            ref: {1,2}
                1 -> Restricted 
                2 -> Unrestricted
            pbs: str. default: "same" (as calculation)
                Basis set for expressing inverted potential
            """
            self.mol = molecule
            self.basis_str = basis
            self.pbs_str = pbs
            self.ref = ref

            if pbs != 'same': # Builds additional mole for secondary basis set
                self.pbs = gto.Mole()
                self.pbs.atom = self.mol.atom
                self.pbs.basis = self.pbs_str
                self.pbs.build()
            else: 
                self.pbs = None

            self.nalpha = self.mol.nelec[0]
            self.nbeta = self.mol.nelec[1]

        def initialize(self):
            """
            Initializes different components for calculation.
            """
            self.nbf = self.mol.nao_nr()
            if self.pbs_str == 'same':
                self.npbs = self.nbf
            else:
                self.npbs = self.pbs.nao_nr()

            self.grid = PySCFGrider(self.mol, self.pbs)
        
        def get_T(self):
            """
            Generates Kinetic Operator in AO basis.
            
            Returns
            -------
            T: np.ndarray. Shape: (nbf, nbf)
            """
            return self.mol.intor('int1e_kin')

        def get_Tpbas(self):
            """
            Generates Kinetic Operator in AO basis for additional basis. 

            Returns
            -------
            T_pbas: np.ndarray. Shape: (nbf, nbf)
            """
            return self.pbs.intor('int1e_kin')

        def get_V(self):
            """
            Generates External Potential in AO basis

            Returns
            -------
            V: np.ndarray. Shape: (nbf, nbf)
            """
            return self.mol.intor('int1e_nuc')

        def get_A(self):
            """
            Generates S^(-0.5)

            Returns
            -------
            A: np.ndarray. Shape: (nbf, nbf)
            """
            A = self.mol.intor('int1e_ovlp')
            A = scipy.linalg.fractional_matrix_power(A, -.5)
            return A

        def get_S(self):
            """
            Builds Overlap matrix of AO basis

            Returns
            -------
            S: np.ndarray. Shape: (nbf, nbf)
            """
            return self.mol.intor('int1e_ovlp')

        def get_S3(self):
            """
            Builds 3 Overlap Matrix. 
            Manually built since Pyscf does not support it. 

            Returns
            -------
            S3: np.ndarray. Shape: (nbf, nbf, nbf or npbs)
                Third dimension depends on wether an additional basis is used. 
            """

            grid = dft.gen_grid.Grids(self.mol)
            grid.build()
            bs1 = dft.numint.eval_ao(self.mol, grid.coords)

            if self.pbs_str == 'same':
                S3 = contract('ij, ik, il, i -> jkl', bs1, bs1, bs1, grid.weights)
                del bs1

            else:
                bs2 = dft.numint.eval_ao(self.pbs, grid.coords)
                S3 = contract('ij, ik, il, i -> jkl', bs1, bs1, bs2, grid.weights)
                del bs1
                del bs2

            return S3

        def get_S4(self):
            """
            Obtains a 4 AO Overlap Matrix using Density Fitting.
            """
            
            grid = dft.gen_grid.Grids(self.mol)
            grid.build()
            bs1 = dft.numint.eval_ao(self.mol, grid.coords)

            # PySCF's DF basis are identified by the suffix "-jk-fit".
            auxbasis = self.mol.basis + '-jk-fit'
            mol_aux = gto.M(atom=self.mol.atom, basis=auxbasis)
            bs2 = dft.numint.eval_ao(mol_aux, grid.coords)

            S_Pmn = contract('ij, ik, il, i -> jkl', bs2, bs1, bs1, grid.weights)
            S_PQ = mol_aux.mol.intor('int1e_ovlp')
            S_PQinv = np.linalg.pinv(S_PQ, rcond=1e-9)
            S4 = contract('Pmn,PQ,Qrs->mnrs', S_Pmn, S_PQinv, S_Pmn)

            return S4

        def compute_hartree(self, Cocc_a, Cocc_b=None):
            """
            Computes Hartree Operator in AO basis

            Parameters
            ----------
            ca: np.ndarray
                Occupied Orbitals in AO basis
            cb: np.ndarray
                if ref == 2, cb -> Beta Occupied Orbitals in AO basis
            """
            da = (Cocc_a @ Cocc_a.T)
            if Cocc_a is not None:
                db = (Cocc_b @ Cocc_b.T)
            else:
                db = da
            mf = dft.uks.UKS(self.mol)
            J = mf.get_j(dm=[da, db])

            return J
        
        def run_single_point(self, mol, basis, method):
            """
            Run a Standard calculation
            """

            # DFT
            molecule = gto.Mole() 
            molecule.atom = mol.atom
            molecule.basis = basis
            molecule.build()        

            if self.ref == 1:
                mf    = dft.RKS(mol)
            else:
                mf    = dft.UKS(mol)
            mf.xc = method
            mf.kernel()

            return mf.make_rdm1(), mf.mo_coeff, mf.mo_energy 


            # Post-SCF
            
        def diagonalize( self, matrix, ndocc ):
            """
            Diagonalizes Fock Matrix

            Parameters
            ----------
            marrix: np.ndarray
                Matrix to be diagonalized
            ndocc: int
                Number of occupied orbitals

            Returns
            -------
            C: np.ndarray
                Orbital Matrix
            Cocc: np.ndarray
                Occupied Orbital Matrix
            D: np.ndarray
                Density Matrix
            eigves: np.ndarray
                Eigenvalues
            """

            self.A = self.get_A()

            Fp = self.A.dot(matrix).dot(self.A)
            eigvecs, Cp = np.linalg.eigh(Fp)
            C = self.A.dot(Cp)
            Cocc = C[:, :ndocc]
            D = contract('pi,qi->pq', Cocc, Cocc)
            return C, Cocc, D, eigvecs
        